@@ -1,4 +1,4 @@
-# Copyright (C) 2010 Kristian B. Oelgaard
+# Copyright (C) 2013 Kristian B. Oelgaard
 #
 # This file is part of FFC.
 #
@@ -15,11 +15,14 @@
 # You should have received a copy of the GNU Lesser General Public License
 # along with FFC. If not, see <http://www.gnu.org/licenses/>.
 #
+# Modified by Marie E. Rognes, 2013
+# Modified by Martin Alnaes, 2013
+#
 # First added:  2010-02-08
-# Last changed: 2011-01-11
+# Last changed: 2013-02-10
 
 # FFC modules
-from ffc.log import info
+from ffc.log import info, error
 from ffc.cpp import format
 from ffc.quadrature.symbolics import optimise_code, BASIS, IP, GEO, CONST
 from ffc.quadrature.symbolics import create_product, create_sum, create_symbol, create_fraction
@@ -35,6 +38,7 @@ def optimize_integral_ir(ir, parameters):
         integrals  =  ir["trans_integrals"]
         domain_type = ir["domain_type"]
         num_facets =  ir["num_facets"]
+        num_vertices =  ir["num_vertices"]
         geo_consts =  ir["geo_consts"]
         psi_tables_map =  ir["psi_tables_map"]
         if domain_type == "cell":
@@ -58,6 +62,13 @@ def optimize_integral_ir(ir, parameters):
                         _precompute_expressions(integrals[i][j], geo_consts,parameters["optimisation"])
                     else:
                         _simplify_expression(integrals[i][j], geo_consts, psi_tables_map)
+        elif domain_type == "point":
+            for i in range(num_vertices):
+                info("Optimising expressions for poin integral %d" % i)
+                if parameters["optimisation"] in ("precompute_ip_const", "precompute_basis_const"):
+                    _precompute_expressions(integrals[i], geo_consts, parameters["optimisation"])
+                else:
+                    _simplify_expression(integrals[i], geo_consts, psi_tables_map)
         else:
             error("Unhandled domain type: " + str(domain_type))
 
@@ -227,4 +238,3 @@ def _reduce_expression(expr, symbols, const_dict, f_name, use_expr_type=False):
         s = create_symbol(f_name(const_dict[new_sym]), sym_type)
         return [s]
     return symbols
-
