@@ -49,8 +49,8 @@ from ffc.log import debug, error
 from ffc.fiatinterface import reference_cell
 from ffc.fiatinterface import create_quadrature as fiat_create_quadrature
 
-# Dictionary mapping from domain (cell) to dimension
-from ufl.geometry import domain2dim
+# Dictionary mapping from cellname to dimension
+from ufl.geometry import cellname2dim
 
 def create_quadrature(shape, degree, scheme="default"):
     """
@@ -60,7 +60,7 @@ def create_quadrature(shape, degree, scheme="default"):
 
     # FIXME: KBO: Can this be handled more elegantly?
     # Handle point case
-    if isinstance(shape, int) and shape == 0 or domain2dim[shape] == 0:
+    if isinstance(shape, int) and shape == 0 or cellname2dim[shape] == 0:
         return ([()], array([1.0,]))
 
     if scheme == "default":
