@@ -12,7 +12,7 @@ function "foo", one should only need to use the data stored
 in the intermediate representation under the key "foo".
 """
 
-# Copyright (C) 2009-2010 Anders Logg
+# Copyright (C) 2009-2013 Anders Logg
 #
 # This file is part of FFC.
 #
@@ -31,32 +31,45 @@ in the intermediate representation under the key "foo".
 #
 # Modified by Marie E. Rognes 2010
 # Modified by Kristian B. Oelgaard 2010
+# Modified by Martin Alnaes, 2013
 #
 # First added:  2009-12-16
-# Last changed: 2010-04-12
+# Last changed: 2013-02-10
 
 # Python modules
 from itertools import chain
 
 # Import UFL
 import ufl
+from ufl.classes import Measure
 
 # FFC modules
 from ffc.utils import compute_permutations, product
 from ffc.log import info, error, begin, end, debug_ir, ffc_assert, warning
-from ffc.fiatinterface import create_element, entities_per_dim, reference_cell
+from ffc.fiatinterface import create_element, cellname_to_num_entities, reference_cell
 from ffc.mixedelement import MixedElement
 from ffc.enrichedelement import EnrichedElement, SpaceOfReals
 from ffc.quadratureelement import QuadratureElement
 from ffc.cpp import set_float_formatting
 
-# FFC specialized representation modules
-from ffc import quadrature
-from ffc import tensor
+def pick_representation(representation):
+    "Return one of the specialized code generation modules from a representation string."
+    if representation == "quadrature":
+        from ffc import quadrature
+        r = quadrature
+    elif representation == "tensor":
+        from ffc import tensor
+        r = tensor
+    elif representation == "uflacs":
+        from ffc import uflacsrepr
+        r = uflacsrepr
+    else:
+        error("Unknown representation: %s" % str(representation))
+    return r
 
 not_implemented = None
 
-def compute_ir(analysis, parameters, common_cell=None):
+def compute_ir(analysis, parameters):
     "Compute intermediate representation."
 
     begin("Compiler stage 2: Computing intermediate representation")
@@ -79,7 +92,7 @@ def compute_ir(analysis, parameters, common_cell=None):
 
     # Compute and flatten representation of integrals
     info("Computing representation of integrals")
-    irs = [_compute_integral_ir(fd, i, parameters, common_cell) \
+    irs = [_compute_integral_ir(fd, i, parameters) \
                for (i, fd) in enumerate(form_datas)]
     ir_integrals = [ir for ir in chain(*irs) if not ir is None]
 
@@ -104,14 +117,14 @@ def _compute_element_ir(ufl_element, element_id, element_numbers):
 
     # Compute data for each function
     ir["signature"] = repr(ufl_element)
-    ir["cell_shape"] = cell.domain()
+    ir["cell_shape"] = cell.cellname()
     ir["topological_dimension"] = cell.topological_dimension()
     ir["geometric_dimension"] = cell.geometric_dimension()
     ir["space_dimension"] = element.space_dimension()
     ir["value_rank"] = len(ufl_element.value_shape())
     ir["value_dimension"] = ufl_element.value_shape()
-    ir["evaluate_basis"] = _evaluate_basis(element, cell)
-    ir["evaluate_dof"] = _evaluate_dof(element, cell)
+    ir["evaluate_basis"] = _evaluate_basis(ufl_element, element, cell)
+    ir["evaluate_dof"] = _evaluate_dof(ufl_element, element, cell)
     ir["interpolate_vertex_values"] = _interpolate_vertex_values(element, cell)
     ir["num_sub_elements"] = ufl_element.num_sub_elements()
     ir["create_sub_element"] = _create_sub_foo(ufl_element, element_numbers)
@@ -137,12 +150,9 @@ def _compute_dofmap_ir(ufl_element, element_id, element_numbers):
     # Compute data for each function
     ir["signature"] = "FFC dofmap for " + repr(ufl_element)
     ir["needs_mesh_entities"] = _needs_mesh_entities(element)
-    ir["init_mesh"] = _init_mesh(element)
-    ir["init_cell"] = None
-    ir["init_cell_finalize"] = None
     ir["topological_dimension"] = cell.topological_dimension()
     ir["geometric_dimension"] = cell.geometric_dimension()
-    ir["global_dimension"] = None
+    ir["global_dimension"] = _global_dimension(element)
     ir["local_dimension"] = element.space_dimension()
     ir["max_local_dimension"] = element.space_dimension()
     ir["num_facet_dofs"] = len(facet_dofs[0])
@@ -150,7 +160,7 @@ def _compute_dofmap_ir(ufl_element, element_id, element_numbers):
     ir["tabulate_dofs"] = _tabulate_dofs(element, cell)
     ir["tabulate_facet_dofs"] = facet_dofs
     ir["tabulate_entity_dofs"] = (element.entity_dofs(), num_dofs_per_entity)
-    ir["tabulate_coordinates"] = _tabulate_coordinates(element)
+    ir["tabulate_coordinates"] = _tabulate_coordinates(ufl_element, element)
     ir["num_sub_dofmaps"] = ufl_element.num_sub_elements()
     ir["create_sub_dofmap"] = _create_sub_foo(ufl_element, element_numbers)
 
@@ -158,7 +168,8 @@ def _compute_dofmap_ir(ufl_element, element_id, element_numbers):
 
     return ir
 
-def _init_mesh(element):
+def _global_dimension(element):
+    "Compute intermediate representation for global_dimension."
 
     if not isinstance(element, MixedElement):
         if isinstance(element, SpaceOfReals):
@@ -176,10 +187,10 @@ def _init_mesh(element):
     element = MixedElement(elements)
     return (_num_dofs_per_entity(element), num_reals)
 
-
 def _needs_mesh_entities(element):
+    "Compute intermediate representation for needs_mesh_entities."
 
-    # Note: The dof map for Real elements does not depend on the mesh 
+    # Note: The dof map for Real elements does not depend on the mesh
 
     num_dofs_per_entity = _num_dofs_per_entity(element)
     if isinstance(element, SpaceOfReals):
@@ -187,31 +198,23 @@ def _needs_mesh_entities(element):
     else:
         return [d > 0 for d in num_dofs_per_entity]
 
-def _compute_integral_ir(form_data, form_id, parameters, common_cell=None):
-    "Compute intermediate represention of form integrals."
+def _compute_integral_ir(form_data, form_id, parameters):
+    "Compute intermediate represention for form integrals."
 
     irs = []
 
     # Iterate over integrals
-    for (domain_type, domain_id, integrals, metadata) in form_data.integral_data:
+    for itg_data in form_data.integral_data:
 
         # Select representation
-        if metadata["representation"] == "quadrature":
-            r = quadrature
-        elif metadata["representation"] == "tensor":
-            r = tensor
-        else:
-            error("Unknown representation: " + str(metadata["representation"]))
+        # TODO: Is it possible to detach this metadata from IntegralData? It's a bit strange from the ufl side.
+        r = pick_representation(itg_data.metadata["representation"])
 
         # Compute representation
-        ir = r.compute_integral_ir(domain_type,
-                                   domain_id,
-                                   integrals,
-                                   metadata,
+        ir = r.compute_integral_ir(itg_data,
                                    form_data,
                                    form_id,
-                                   parameters,
-                                   common_cell)
+                                   parameters)
 
         # Append representation
         irs.append(ir)
@@ -224,25 +227,32 @@ def _compute_form_ir(form_data, form_id, element_numbers):
     # Store id
     ir = {"id": form_id}
 
-    #print "check 2", element_numbers
-    #print [element_numbers[e] for e in form_data.elements]
-
     # Compute common data
     ir["classname"] = "FooForm"
     ir["members"] = not_implemented
     ir["constructor"] = not_implemented
     ir["destructor"] = not_implemented
-    ir["signature"] = repr(form_data.preprocessed_form)
+    ir["signature"] = form_data.signature
     ir["rank"] = form_data.rank
     ir["num_coefficients"] = form_data.num_coefficients
-    ir["num_cell_domains"] = form_data.num_cell_domains
-    ir["num_exterior_facet_domains"] = form_data.num_exterior_facet_domains
-    ir["num_interior_facet_domains"] = form_data.num_interior_facet_domains
+    ir["num_cell_domains"] = form_data.num_sub_domains.get("cell",0)
+    ir["num_exterior_facet_domains"] = form_data.num_sub_domains.get("exterior_facet",0)
+    ir["num_interior_facet_domains"] = form_data.num_sub_domains.get("interior_facet",0)
+    ir["num_point_domains"] = form_data.num_sub_domains.get("point",0)
+    ir["has_cell_integrals"] = _has_foo_integrals("cell", form_data)
+    ir["has_exterior_facet_integrals"] = _has_foo_integrals("exterior_facet", form_data)
+    ir["has_interior_facet_integrals"] = _has_foo_integrals("interior_facet", form_data)
+    ir["has_point_integrals"] = _has_foo_integrals("point", form_data)
     ir["create_finite_element"] = [element_numbers[e] for e in form_data.elements]
     ir["create_dofmap"] = [element_numbers[e] for e in form_data.elements]
     ir["create_cell_integral"] = _create_foo_integral("cell", form_data)
     ir["create_exterior_facet_integral"] = _create_foo_integral("exterior_facet", form_data)
     ir["create_interior_facet_integral"] = _create_foo_integral("interior_facet", form_data)
+    ir["create_point_integral"] = _create_foo_integral("point", form_data)
+    ir["create_default_cell_integral"] = _create_default_foo_integral("cell", form_data)
+    ir["create_default_exterior_facet_integral"] = _create_default_foo_integral("exterior_facet", form_data)
+    ir["create_default_interior_facet_integral"] = _create_default_foo_integral("interior_facet", form_data)
+    ir["create_default_point_integral"] = _create_default_foo_integral("point", form_data)
 
     return ir
 
@@ -250,42 +260,81 @@ def _compute_form_ir(form_data, form_id, element_numbers):
 
 # FIXME: Move to FiniteElement/MixedElement
 def _value_size(element):
-    "Compute value size of element."
+    """Compute value size of element, aka the number of components.
+
+    The value size of a scalar field is 1, the value size of a vector
+    field (is the number of components), the value size of a higher
+    dimensional tensor field is the product of the value_shape of the
+    field. Recall that all mixed elements are flattened.
+    """
     shape = element.value_shape()
     if shape == ():
         return 1
-    else:
-        # FIXME: KBO: Is this correct? Shouldn't it be sum() instead?
-        return product(shape)
-
-def _generate_offsets(element, offset=0):
-
-    "Generate offsets: i.e value offset for each basis function."
+    return product(shape)
 
+def _generate_reference_offsets(element, offset=0):
+    """Generate offsets: i.e value offset for each basis function
+    relative to a reference element representation."""
     offsets = []
-
     if isinstance(element, MixedElement):
         for e in element.elements():
-            offsets += _generate_offsets(e, offset)
+            offsets += _generate_reference_offsets(e, offset)
             offset += _value_size(e)
-
     elif isinstance(element, EnrichedElement):
         for e in element.elements():
-            offsets += _generate_offsets(e, offset)
-
+            offsets += _generate_reference_offsets(e, offset)
     else:
         offsets = [offset]*element.space_dimension()
-
     return offsets
 
-def _evaluate_dof(element, cell):
+def _generate_physical_offsets(ufl_element, offset=0):
+    """Generate offsets: i.e value offset for each basis function
+    relative to a physical element representation."""
+    offsets = []
+
+    # Refer to reference if gdim == tdim. This is a hack to support
+    # more stuff (in particular restricted elements)
+    gdim = ufl_element.cell().geometric_dimension()
+    tdim = ufl_element.cell().topological_dimension()
+    if (gdim == tdim):
+        return _generate_reference_offsets(create_element(ufl_element))
+
+    if isinstance(ufl_element, ufl.MixedElement):
+        for e in ufl_element.sub_elements():
+            offsets += _generate_physical_offsets(e, offset)
+            offset += _value_size(e)
+    elif isinstance(ufl_element, ufl.EnrichedElement):
+        for e in ufl_element._elements:
+            offsets += _generate_physical_offsets(e, offset)
+    elif isinstance(ufl_element, ufl.FiniteElement):
+        element = create_element(ufl_element)
+        offsets = [offset]*element.space_dimension()
+    else:
+        raise NotImplementedError, \
+            "This element combination is not implemented"
+    return offsets
+
+def _evaluate_dof(ufl_element, element, cell):
     "Compute intermediate representation of evaluate_dof."
 
+    # With regard to reference_value_size vs physical_value_size: Note
+    # that 'element' is the FFC/FIAT representation of the finite
+    # element, while 'ufl_element' is the UFL representation. In
+    # particular, UFL only knows about physical dimensions, so the
+    # value shape of the 'ufl_element' (which is used to compute the
+    # _value_size) will be correspond to the value size in physical
+    # space. FIAT however only knows about the reference element, and
+    # so the FIAT value shape of the 'element' will be the reference
+    # value size. This of course only matters for elements that have
+    # different physical and reference value shapes and sizes.
+
     return {"mappings": element.mapping(),
-            "value_size": _value_size(element),
-            "cell_dimension": cell.geometric_dimension(),
+            "reference_value_size": _value_size(element),
+            "physical_value_size": _value_size(ufl_element),
+            "geometric_dimension": cell.geometric_dimension(),
+            "topological_dimension": cell.topological_dimension(),
             "dofs": [L.pt_dict for L in element.dual_basis()],
-            "offsets": _generate_offsets(element)}
+            "physical_offsets": _generate_physical_offsets(ufl_element)}
 
 def _extract_elements(element):
 
@@ -297,25 +346,30 @@ def _extract_elements(element):
         new_elements.append(element)
     return new_elements
 
-def _num_components(element):
-    """Compute the number of components of element, like _value_size, but
-    does not support tensor elements."""
-    shape = element.value_shape()
-    if shape == ():
-        return 1
-    elif len(shape) == 1:
-        return shape[0]
-    else:
-        error("Tensor valued elements are not supported yet: %d " % shape)
+# def _num_components(element):
+#     """Compute the number of components of element, like _value_size, but
+#     does not support tensor elements."""
+#     shape = element.value_shape()
+#     if shape == ():
+#         return 1
+#     elif len(shape) == 1:
+#         return shape[0]
+#     else:
+#         error("Tensor valued elements are not supported yet: %d " % shape)
 
-def _evaluate_basis(element, cell):
+def _evaluate_basis(ufl_element, element, cell):
     "Compute intermediate representation for evaluate_basis."
 
     # Handle Mixed and EnrichedElements by extracting 'sub' elements.
     elements = _extract_elements(element)
-    offsets = _generate_offsets(element)
+    offsets = _generate_reference_offsets(element) # Must check?
     mappings = element.mapping()
 
+    # This function is evidently not implemented for TensorElements
+    for e in elements:
+        if len(e.value_shape()) > 1:
+            return "Function not supported/implemented for TensorElements."
+
     # Handle QuadratureElement, not supported because the basis is only defined
     # at the dof coordinates where the value is 1, so not very interesting.
     for e in elements:
@@ -323,19 +377,22 @@ def _evaluate_basis(element, cell):
             return "Function not supported/implemented for QuadratureElement."
 
     # Initialise data with 'global' values.
-    data = {
-          "value_size" : sum(element.value_shape() or (1,)),
-          "cell_domain" : cell.domain(),
-          "topological_dimension" : cell.topological_dimension(),
-          "geometric_dimension" : cell.geometric_dimension(),
-          "space_dimension" : element.space_dimension()
-          }
+    data = {"reference_value_size": _value_size(element),
+            "physical_value_size": _value_size(ufl_element),
+            "cellname" : cell.cellname(),
+            "topological_dimension" : cell.topological_dimension(),
+            "geometric_dimension" : cell.geometric_dimension(),
+            "space_dimension" : element.space_dimension(),
+            "needs_oriented": needs_oriented_jacobian(element)
+            }
+
+
     # Loop element and space dimensions to generate dof data.
     dof = 0
     dof_data = []
     for e in elements:
         for i in range(e.space_dimension()):
-            num_components = _num_components(e)
+            num_components = _value_size(e)
             coefficients = []
             coeffs = e.get_coeffs()
 
@@ -364,11 +421,17 @@ def _evaluate_basis(element, cell):
 
     return data
 
-def _tabulate_coordinates(element):
+def _tabulate_coordinates(ufl_element, element):
     "Compute intermediate representation of tabulate_coordinates."
+
     if uses_integral_moments(element):
-        return None
-    return [L.pt_dict.keys()[0] for L in element.dual_basis()]
+        return {}
+
+    data = {}
+    data["tdim"] = ufl_element.cell().topological_dimension()
+    data["gdim"] = ufl_element.cell().geometric_dimension()
+    data["points"] = [L.pt_dict.keys()[0] for L in element.dual_basis()]
+    return data
 
 def _tabulate_dofs(element, cell):
     "Compute intermediate representation of tabulate_dofs."
@@ -376,8 +439,8 @@ def _tabulate_dofs(element, cell):
     if isinstance(element, SpaceOfReals):
         return None
 
-    # Extract number of enties for each dimension for this cell
-    num_entities = entities_per_dim[cell.geometric_dimension()]
+    # Extract number of entities for each dimension for this cell
+    num_entities = cellname_to_num_entities[cell.cellname()]
 
     # Extract number of dofs per entity for each element
     elements = all_elements(element)
@@ -407,16 +470,16 @@ def _tabulate_facet_dofs(element, cell):
     "Compute intermediate representation of tabulate_facet_dofs."
 
     # Compute incidences
-    incidence = __compute_incidence(cell.geometric_dimension())
+    incidence = __compute_incidence(cell.topological_dimension())
 
     # Get topological dimension
     D = max([pair[0][0] for pair in incidence])
 
     # Get the number of facets
-    num_facets = cell.num_facets()
+    num_facets = cellname_to_num_entities[cell.cellname()][-2]
 
     # Find out which entities are incident to each facet
-    incident = num_facets*[[]]
+    incident = num_facets*[None]
     for facet in range(num_facets):
         incident[facet] = [pair[1] for pair in incidence if incidence[pair] == True and pair[0] == (D - 1, facet)]
 
@@ -442,14 +505,16 @@ def _interpolate_vertex_values(element, cell):
             return "Function is not supported/implemented for QuadratureElement."
 
     ir = {}
-    ir["cell_dim"] = cell.geometric_dimension()
+    ir["geometric_dimension"] = cell.geometric_dimension()
+    ir["topological_dimension"] = cell.topological_dimension()
 
     # Check whether computing the Jacobian is necessary
     mappings = element.mapping()
     ir["needs_jacobian"] = any("piola" in m for m in mappings)
+    ir["needs_oriented"] = needs_oriented_jacobian(element)
 
     # Get vertices of reference cell
-    cell = reference_cell(cell.domain())
+    cell = reference_cell(cell.cellname())
     vertices = cell.get_vertices()
 
     # Compute data for each constituent element
@@ -467,12 +532,28 @@ def _create_sub_foo(ufl_element, element_numbers):
 
 def _create_foo_integral(domain_type, form_data):
     "Compute intermediate representation of create_foo_integral."
-    return [domain_id for (_domain_type, domain_id, integrals, metadata) in
-           form_data.integral_data if _domain_type == domain_type]
+    return [itg_data.domain_id for itg_data in form_data.integral_data
+           if itg_data.domain_type == domain_type and isinstance(itg_data.domain_id, int)]
+
+def _has_foo_integrals(domain_type, form_data):
+    "Compute intermediate representation of has_foo_integrals."
+    v = (form_data.num_sub_domains.get(domain_type,0) > 0
+         or _create_default_foo_integral(domain_type, form_data) is not None)
+    return bool(v)
+
+def _create_default_foo_integral(domain_type, form_data):
+    "Compute intermediate representation of create_default_foo_integral."
+    itg_data = [itg_data for itg_data in form_data.integral_data
+           if itg_data.domain_id == Measure.DOMAIN_ID_OTHERWISE and itg_data.domain_type == domain_type]
+    ffc_assert(len(itg_data) in (0,1), "Expecting at most one default integral of each type.")
+    return Measure.DOMAIN_ID_OTHERWISE if itg_data else None
 
 #--- Utility functions ---
 
 # FIXME: KBO: This could go somewhere else, like in UFL?
+#        MSA: There is probably something related in ufl somewhere,
+#        but I don't understand quite what this does.
+#        In particular it does not cover sub-sub-elements? Is that a bug?
 # Also look at function naming, use single '_' for utility functions.
 def all_elements(element):
 
@@ -514,7 +595,8 @@ def __compute_incidence(D):
     return incidence
 
 def __compute_sub_simplices(D, d):
-    "Compute vertices for all sub simplices of dimension d (code taken from Exterior)"
+    """Compute vertices for all sub simplices of dimension d (code
+    taken from Exterior)."""
 
     # Number of vertices
     num_vertices = D + 1
@@ -542,7 +624,13 @@ def __compute_sub_simplices(D, d):
     return sub_simplices
 
 def uses_integral_moments(element):
+    "True if element uses integral moments for its degrees of freedom."
 
     integrals = set(["IntegralMoment", "FrobeniusIntegralMoment"])
     tags = set([L.get_type_tag() for L in element.dual_basis()])
     return len(integrals & tags) > 0
+
+def needs_oriented_jacobian(element):
+    # Check whether this element needs an oriented jacobian (only
+    # contravariant piolas seem to need it)
+    return ("contravariant piola" in element.mapping())
