@@ -9,7 +9,7 @@ form files found in the 'bench' directory. To run benchmarks, use the
 option --bench.
 """
 
-# Copyright (C) 2010 Anders Logg, Kristian B. Oelgaard and Marie E. Rognes
+# Copyright (C) 2010-2013 Anders Logg, Kristian B. Oelgaard and Marie E. Rognes
 #
 # This file is part of FFC.
 #
@@ -26,8 +26,10 @@ option --bench.
 # You should have received a copy of the GNU Lesser General Public License
 # along with FFC. If not, see <http://www.gnu.org/licenses/>.
 #
+# Modified by Martin Alnaes, 2013
+#
 # First added:  2010-01-21
-# Last changed: 2011-11-28
+# Last changed: 2013-02-14
 
 # FIXME: Need to add many more test cases. Quite a few DOLFIN forms
 # failed after the FFC tests passed.
@@ -37,6 +39,7 @@ from numpy import array, shape, abs, max, isnan
 from ffc.log import begin, end, info, info_red, info_green, info_blue
 from ufctest import generate_test_code
 from instant.output import get_status_output
+import time
 
 # Parameters
 output_tolerance = 1.e-6
@@ -56,9 +59,19 @@ ext_quad = [\
 "-r quadrature -O -fprecompute_basis_const -feliminate_zeros",
 ]
 
+# Extended uflacs tests (to be extended with optimisation parameters later)
+ext_uflacs = [\
+"-r uflacs",
+]
+
+_command_timings = []
 def run_command(command):
     "Run command and collect errors in log file."
+    t1 = time.time()
     (status, output) = get_status_output(command)
+    t2 = time.time()
+    global _command_timings
+    _command_timings.append((command, t2-t1))
     if status == 0:
         return True
     global logfile
@@ -81,7 +94,7 @@ def clean_output(output_directory):
         shutil.rmtree(output_directory)
     os.mkdir(output_directory)
 
-def generate_test_cases(bench):
+def generate_test_cases(bench, only_forms):
     "Generate form files for all test cases."
 
     begin("Generating test cases")
@@ -91,8 +104,13 @@ def generate_test_cases(bench):
         form_directory = bench_directory
     else:
         form_directory = demo_directory
+
+    # Make list of form files
     form_files = [f for f in os.listdir(form_directory) if f.endswith(".ufl")]
+    if only_forms:
+        form_files = [f for f in form_files if f in only_forms]
     form_files.sort()
+
     for f in form_files:
         shutil.copy("%s/%s" % (form_directory, f), ".")
     info_green("Found %d form files" % len(form_files))
@@ -109,20 +127,28 @@ def generate_test_cases(bench):
 
     end()
 
-def generate_code(args):
+def generate_code(args, only_forms):
     "Generate code for all test cases."
 
     # Get a list of all files
     form_files = [f for f in os.listdir(".") if f.endswith(".ufl")]
+    if only_forms:
+        form_files = [f for f in form_files if f in only_forms]
     form_files.sort()
 
     begin("Generating code (%d form files found)" % len(form_files))
 
+    # TODO: Parse additional options from .ufl file? I.e. grep for some sort of tag like '#ffc: <flags>'.
+    special = {
+        "AdaptivePoisson.ufl": "-e",
+        }
+
     # Iterate over all files
     for f in form_files:
+        options = special.get(f, "")
 
-        cmd = ("ffc %s -f precision=8 -fconvert_exceptions_to_warnings %s"
-               % (" ".join(args), f))
+        cmd = ("ffc %s %s -f precision=8 -fconvert_exceptions_to_warnings %s"
+               % (options, " ".join(args), f))
 
         # Generate code
         ok = run_command(cmd)
@@ -164,14 +190,14 @@ def validate_code(reference_dir):
         else:
             info_red("%s differs" % f)
             diff = "\n".join([line for line in difflib.unified_diff(reference_code.split("\n"), generated_code.split("\n"))])
-            s = ("Code differs for %s, diff follows"
+            s = ("Code differs for %s, diff follows (reference first, generated second)"
                  % os.path.join(*reference_file.split(os.path.sep)[-3:]))
             log_error("\n" + s + "\n" + len(s)*"-")
             log_error(diff)
 
     end()
 
-def build_programs(bench):
+def build_programs(bench, permissive):
     "Build test programs for all test cases."
 
     # Get a list of all files
@@ -194,6 +220,7 @@ def build_programs(bench):
     # If BOOST_DIR is not set use default directory
     boost_inc_dir = ""
     boost_lib_dir = ""
+    boost_math_tr1_lib = "boost_math_tr1"
     boost_dir = os.getenv("BOOST_DIR", default)
     boost_is_found = False
     for inc_dir in ["", "include"]:
@@ -201,10 +228,13 @@ def build_programs(bench):
             boost_inc_dir = os.path.join(boost_dir, inc_dir)
             break
     for lib_dir in ["", "lib"]:
-        if os.path.isfile(os.path.join(boost_dir, lib_dir, "libboost_math_tr1.so")) or\
-           os.path.isfile(os.path.join(boost_dir, lib_dir, "libboost_math_tr1.dylib")):
-            boost_lib_dir = os.path.join(boost_dir, lib_dir)
-            break
+        for ext in [".so", "-mt.so", ".dylib", "-mt.dylib"]:
+            _lib = os.path.join(boost_dir, lib_dir, "lib" + boost_math_tr1_lib + ext)
+            if os.path.isfile(_lib):
+                if "-mt" in _lib:
+                    boost_math_tr1_lib += "-mt"
+                boost_lib_dir = os.path.join(boost_dir, lib_dir)
+                break
     if boost_inc_dir != "" and boost_lib_dir != "":
         boost_is_found = True
 
@@ -217,24 +247,25 @@ set the environment variable BOOST_DIR.
     ufc_cflags += " -I%s -L%s" % (boost_inc_dir, boost_lib_dir)
 
     # Set compiler options
-    if bench > 0:
+    compiler_options = "%s -Wall" % ufc_cflags
+    if not permissive:
+        compiler_options += " -Werror"
+    if bench:
         info("Benchmarking activated")
         # Takes too long to build with -O2
-        #compiler_options = "%s -Wall -Werror -O2" % ufc_cflags
-        compiler_options = "%s -Wall -Werror" % ufc_cflags
-    else:
-        compiler_options = "%s -Wall -Werror -g" % ufc_cflags
+        #compiler_options += " -O2"
     info("Compiler options: %s" % compiler_options)
 
     # Iterate over all files
     for f in header_files:
 
         # Generate test code
-        filename = generate_test_code(f, bench)
+        filename = generate_test_code(f)
 
         # Compile test code
         prefix = f.split(".h")[0]
-        command = "g++ %s -o %s.bin %s.cpp -lboost_math_tr1" % (compiler_options, prefix, prefix)
+        command = "g++ %s -o %s.bin %s.cpp -l%s" % \
+                  (compiler_options, prefix, prefix, boost_math_tr1_lib)
         ok = run_command(command)
 
         # Check status
@@ -245,9 +276,12 @@ set the environment variable BOOST_DIR.
 
     end()
 
-def run_programs():
+def run_programs(bench):
     "Run generated programs."
 
+    # This matches argument parsing in the generated main files
+    bench = 'b' if bench else ''
+
     # Get a list of all files
     test_programs = [f for f in os.listdir(".") if f.endswith(".bin")]
     test_programs.sort()
@@ -263,7 +297,7 @@ def run_programs():
             os.remove(prefix + ".out")
         except:
             pass
-        ok = run_command(".%s%s.bin > %s.out" % (os.path.sep, prefix, prefix))
+        ok = run_command(".%s%s.bin %s > %s.out" % (os.path.sep, prefix, bench, prefix))
 
         # Check status
         if ok:
@@ -300,7 +334,7 @@ def validate_programs(reference_dir):
         ok = True
         old = [line.split(" = ") for line in reference_output.split("\n") if " = " in line]
         new = dict([line.split(" = ") for line in generated_output.split("\n") if " = " in line])
-        header = ("Output differs for %s, diff follows"
+        header = ("Output differs for %s, diff follows (reference first, generated second)"
                   % os.path.join(*reference_file.split(os.path.sep)[-3:]))
         for (key, value) in old:
 
@@ -341,20 +375,86 @@ def validate_programs(reference_dir):
         else:
             info_red("%s differs" % f)
 
+
+        # Now check json references
+        fj = f.replace(".out", ".json")
+
+        # Get generated json output
+        if os.path.exists(fj):
+            generated_json_output = open(fj).read()
+            if "nan" in generated_json_output:
+                info_red("Found nan in generated json output, replacing with 999 to be able to parse as python dict.")
+                generated_json_output = generated_json_output.replace("nan", "999")
+        else:
+            generated_json_output = "{}"
+
+        # Get reference json output
+        reference_json_file = os.path.join(reference_dir, fj)
+        if os.path.isfile(reference_json_file):
+            reference_json_output = open(reference_json_file).read()
+        else:
+            info_blue("Missing reference for %s" % reference_json_file)
+            reference_json_output = "{}"
+
+        # Compare json with reference using recursive diff algorithm # TODO: Write to different error file?
+        from recdiff import recdiff, print_recdiff, DiffEqual
+        # Assuming reference is well formed
+        reference_json_output = eval(reference_json_output)
+        try:
+            generated_json_output = eval(generated_json_output)
+        except Exception as e:
+            info_red("Failed to evaluate json output for %s" % fj)
+            log_error(str(e))
+            generated_json_output = None
+        json_diff = (None if generated_json_output is None else
+                     recdiff(generated_json_output, reference_json_output, tolerance=output_tolerance))
+        json_ok = json_diff == DiffEqual
+
+        # Check status
+        if json_ok:
+            info_green("%s OK" % fj)
+        else:
+            info_red("%s differs" % fj)
+            log_error("Json output differs for %s, diff follows (reference first, generated second)"
+                      % os.path.join(*reference_json_file.split(os.path.sep)[-3:]))
+            print_recdiff(json_diff, printer=log_error)
+
     end()
 
 def main(args):
     "Run all regression tests."
 
-    # Check command-line arguments
-    bench = "--bench" in args
-    fast = "--fast" in args
-    ext = "--ext_quad" in args
-    generate_only = "--generate-only" in args
+    # Check command-line arguments TODO: Use getargs or something
+    generate_only  = "--generate-only" in args
+    fast           = "--fast" in args
+    bench          = "--bench" in args
+    use_quad       = "--skip_quad" not in args
+    use_ext_quad   = "--ext_quad" in args
+    use_ext_uflacs = "--ext_uflacs" in args
+    permissive     = "--permissive" in args
+    tolerant       = "--tolerant" in args
+    print_timing   = "--print-timing" in args
 
-    args = [arg for arg in args
-            if not arg in ("--bench", "--fast", "--ext_quad",
-                           "--generate-only")]
+    flags = (
+        "--generate-only",
+        "--fast",
+        "--bench",
+        "--skip_quad",
+        "--ext_quad",
+        "--ext_uflacs",
+        "--permissive",
+        "--tolerant",
+        "--print-timing",
+        )
+    args = [arg for arg in args if not arg in flags]
+
+    # Extract .ufl names from args
+    only_forms = set([arg for arg in args if arg.endswith(".ufl")])
+    args = [arg for arg in args if arg not in only_forms]
+
+    if tolerant:
+        global output_tolerance
+        output_tolerance = 1e-3
 
     # Clean out old output directory
     output_directory = "output"
@@ -364,10 +464,13 @@ def main(args):
     # Adjust which test cases (combinations of compile arguments) to
     # run here
     test_cases = ["-r auto"]
-    if (not bench and not fast):
+    if use_quad and (not bench and not fast):
         test_cases += ["-r quadrature", "-r quadrature -O"]
-        if ext:
-            test_cases += ext_quad
+    if use_ext_quad:
+        test_cases += ext_quad
+    if use_ext_uflacs:
+        test_cases = ext_uflacs
+        #test_cases += ["-r quadrature -O"]
 
     for argument in test_cases:
 
@@ -379,10 +482,10 @@ def main(args):
         os.chdir(sub_directory)
 
         # Generate test cases
-        generate_test_cases(bench)
+        generate_test_cases(bench, only_forms)
 
         # Generate code
-        generate_code(args + [argument])
+        generate_code(args + [argument], only_forms)
 
         # Location of reference directories
         reference_directory =  os.path.abspath("../../references/")
@@ -399,11 +502,11 @@ def main(args):
         if fast or generate_only:
             info("Skipping program validation")
         elif bench:
-            build_programs(bench)
-            run_programs()
+            build_programs(bench, permissive)
+            run_programs(bench)
         else:
-            build_programs(bench)
-            run_programs()
+            build_programs(bench, permissive)
+            run_programs(bench)
             validate_programs(output_reference_dir)
 
         # Go back up
@@ -412,6 +515,10 @@ def main(args):
         end()
 
     # Print results
+    if print_timing:
+        timings = '\n'.join("%10.2e s  %s" % (t, name) for (name, t) in _command_timings)
+        info_green("Timing of all commands executed:")
+        info(timings)
     if logfile is None:
         info_green("Regression tests OK")
         return 0
